import pytest
from yaml_where.yaml_where import YAMLWhereSequence
from helpers import clean_yaml, rng
from yaml_where import YAMLWhere
from yaml_where.exceptions import MissingKeyError, UndefinedAccessError


def test_get_top_level():
    yaml = """
    [1,
     a, foo,
     
         indented]
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get(0) == rng(0, 1, 0, 2)
    assert source_map.get(1) == rng(1, 1, 1, 2)
    assert source_map.get(2) == rng(1, 4, 1, 7)
    assert source_map.get(3) == rng(3, 5, 3, 13)


def test_get_nested():
    yaml = """
    - [1, 2, 34]
    - [3, [4, 5, 6]]
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get(0) == rng(0, 2, 0, 12)
    assert source_map.get(0, 0) == rng(0, 3, 0, 4)
    assert source_map.get(0, 1) == rng(0, 6, 0, 7)
    assert source_map.get(0, 2) == rng(0, 9, 0, 11)
    assert source_map.get(1, 0) == rng(1, 3, 1, 4)
    assert source_map.get(1, 1) == rng(1, 6, 1, 15)
    assert source_map.get(1, 1, 0) == rng(1, 7, 1, 8)
    assert source_map.get(1, 1, 1) == rng(1, 10, 1, 11)
    assert source_map.get(1, 1, 2) == rng(1, 13, 1, 14)


def test_get_value_top_level():
    yaml = """
    [1,
     a, foo,
     
         indented]
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get_value(0) == rng(0, 1, 0, 2)
    assert source_map.get_value(1) == rng(1, 1, 1, 2)
    assert source_map.get_value(2) == rng(1, 4, 1, 7)
    assert source_map.get_value(3) == rng(3, 5, 3, 13)


def test_get_value_nested():
    yaml = """
    - [1, 2, 34]
    - [3, [4, 5, 6]]
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get_value(0) == rng(0, 2, 0, 12)
    assert source_map.get_value(0, 0) == rng(0, 3, 0, 4)
    assert source_map.get_value(0, 1) == rng(0, 6, 0, 7)
    assert source_map.get_value(0, 2) == rng(0, 9, 0, 11)
    assert source_map.get_value(1, 0) == rng(1, 3, 1, 4)
    assert source_map.get_value(1, 1) == rng(1, 6, 1, 15)
    assert source_map.get_value(1, 1, 0) == rng(1, 7, 1, 8)
    assert source_map.get_value(1, 1, 1) == rng(1, 10, 1, 11)
    assert source_map.get_value(1, 1, 2) == rng(1, 13, 1, 14)


def test_get_key_top_level():
    yaml = "[1, 2, 3]"
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(UndefinedAccessError):
        source_map.get_key(0)


def test_get_missing_index_top_level():
    yaml = "[1, 2, 3]"
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(MissingKeyError):
        source_map.get(4)


def test_get_value_missing_index_top_level():
    yaml = "[1, 2, 3]"
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(MissingKeyError):
        source_map.get_value(4)


def test_get_missing_index_nested():
    yaml = "[1, 2, [3, 4, 5]]"
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(MissingKeyError):
        source_map.get(2, 3)


def test_get_value_missing_index_nested():
    yaml = "[1, 2, [3, 4, 5]]"
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(MissingKeyError):
        source_map.get_value(2, 3)


def test_get_non_integer_key():
    yaml = """
    [1, 2, 3]
    """
    yw = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(UndefinedAccessError):
        yw.get("a")


def test_get_value_non_integer_key():
    yaml = "[1, 2, 3]"
    with pytest.raises(UndefinedAccessError):
        YAMLWhere.from_string(yaml).get_value("a")


def test_constructor_checks_node_type():
    with pytest.raises(ValueError):
        YAMLWhereSequence(None)


"""Main implementation of source map calculation.
"""

from abc import ABC, abstractmethod
from functools import singledispatch

from ruamel.yaml import YAML, MappingNode, Node, ScalarNode, SequenceNode
from yaml_where.exceptions import MissingKeyError, UndefinedAccessError, UnsupportedNodeTypeError
from yaml_where.range import Position, Range


class YAMLWhere(ABC):
    """Base class for source map calculators for different YAML data types.

    This defines the interface for all source map calculators, but is itself abstract.
    """

    @classmethod
    def from_string(cls, source: str):
        """Create a YAMLWhere from a YAML string.

        Args:
            source (str): The YAML string to parse.

        Returns:
            YAMLWhere: The YAMLWhere instance with source map information.
        """
        y = YAML(typ="rt")
        node = y.compose(source)
        return _from_node(node)

    def __init__(self, node: Node):
        self.node = node

    @abstractmethod
    def get(self, *keys: str | int) -> Range:
        """Get the range for an entire entry.

        Different subclasses will have different behaviors for this method. Mappings will return the range
        for the key-value pair, while sequences will return the range for the sequence element.

        Raises:
            MissingKeyError: A key is of the appropriate type for an element, but is missing in that elements. For
                example, if an integer index is beyond the end of a sequence element.
            UndefinedAccessError: If a key is of an inappropriate type for an element. For example, if a string is
                used to access a sequence element.
        """

    @abstractmethod
    def get_key(self, key: str | int, *keys: str | int) -> Range:
        """Get the range for a mapping key.

        So for a mapping, this would return the range for the key in the key-value pair.

        Raises:
            MissingKeyError: A key is of the appropriate type for an element, but is missing in that elements. For
                example, if an integer index is beyond the end of a sequence element.
            UndefinedAccessError: If a key is of an inappropriate type for an element. For example, if a string is
                used to access a sequence element. Or if 'key' is no a defined concept for the element.
        """

    @abstractmethod
    def get_value(self, key: str | int, *keys: str | int) -> Range:
        """Get the range for a value.

        For a mapping this gets the range of the value in the key-value pair. For a sequence this gets the range
        of the sequence element (just like get()).

        Raises:
            MissingKeyError: A key is of the appropriate type for an element, but is missing in that elements. For
                example, if an integer index is beyond the end of a sequence element.
            UndefinedAccessError: If a key is of an inappropriate type for an element. For example, if a string is
                used to access a sequence element. Or if 'key' is no a defined concept for the element.
        """



class YAMLWhereScalar(YAMLWhere):
    "Source map calculator for scalar nodes."

    def __init__(self, node: ScalarNode):
        if not isinstance(node, ScalarNode):
            raise ValueError(f"YAMLWhereScalar can not be constructed with a {type(node).__name__}")
        super().__init__(node)

    def get(self, *keys: str | int) -> Range:
        if keys:
            raise UndefinedAccessError("get() with no arguments is not defined for scalar nodes")

        return Range(
            Position(self.node.start_mark.line, self.node.start_mark.column),
            Position(self.node.end_mark.line, self.node.end_mark.column),
        )

    def get_key(self, key: str | int, *keys: str | int) -> Range:
        raise UndefinedAccessError("get_key() is not defined for scalar nodes")

    def get_value(self, key: str | int, *keys: str | int) -> Range:
        raise UndefinedAccessError("get_value() is not defined for scalar nodes")


class YAMLWhereSequence(YAMLWhere):
    "Source map calculator for sequence nodes."

    def __init__(self, node: SequenceNode):
        if not isinstance(node, SequenceNode):
            raise ValueError(
                f"YAMLWhereSequence can not be constructed with a {type(node).__name__}"
            )

        super().__init__(node)

    def get(self, *keys: str | int) -> Range:
        if not keys:
            raise UndefinedAccessError(
                "get() with no arguments is not defined for sequence elements"
            )

        key, keys = keys[0], keys[1:]

        if not isinstance(key, int):
            raise UndefinedAccessError(f"Can not access a sequence with non-integer key {key}")

        if keys:
            return _from_node(self.node.value[key]).get(*keys)

        try:
            value_node = self.node.value[key]
        except IndexError as err:
            raise MissingKeyError(key) from err

        return Range(
            Position(value_node.start_mark.line, value_node.start_mark.column),
            Position(value_node.end_mark.line, value_node.end_mark.column),
        )

    def get_key(self, key: str | int, *keys: str | int) -> Range:
        if not keys:
            raise UndefinedAccessError("get_key() is not defined for sequence elements")

        return _from_node(self.node.value[key]).get_key(*keys)

    def get_value(self, key: str | int, *keys: str | int) -> Range:
        if not isinstance(key, int):
            raise UndefinedAccessError(f"Can not access a sequence with non-integer key {key}")

        if keys:
            return _from_node(self.node.value[key]).get_value(*keys)

        try:
            value_node = self.node.value[key]
        except IndexError as err:
            raise MissingKeyError(key) from err

        return Range(
            Position(value_node.start_mark.line, value_node.start_mark.column),
            Position(value_node.end_mark.line, value_node.end_mark.column),
        )


class YAMLWhereMapping(YAMLWhere):
    "Source map calculator for mapping nodes."

    def __init__(self, node: MappingNode):
        if not isinstance(node, MappingNode):
            raise ValueError(
                f"YAMLWhereMapping can not be constructed with a {type(node).__name__}"
            )

        super().__init__(node)

    def get(self, *keys: str | int) -> Range:
        if not keys:
            raise UndefinedAccessError(
                "get() with no arguments is not defined for sequence elements"
            )

        key, keys = keys[0], keys[1:]

        for child_key, child_value in self.node.value:
            if child_key.value == key:
                if not keys:
                    return Range(
                        Position(child_key.start_mark.line, child_key.start_mark.column),
                        Position(child_value.end_mark.line, child_value.end_mark.column),
                    )
                else:
                    return _from_node(child_value).get(*keys)

        raise MissingKeyError(key)

    def get_key(self, key: str | int, *keys: str | int) -> Range:
        for child_key, child_value in self.node.value:
            if child_key.value == key:
                if not keys:
                    return Range(
                        Position(child_key.start_mark.line, child_key.start_mark.column),
                        Position(child_key.end_mark.line, child_key.end_mark.column),
                    )
                else:
                    return _from_node(child_value).get_key(*keys)

        raise MissingKeyError(key)

    def get_value(self, key: str | int, *keys: str | int) -> Range:
        for child_key, child_value in self.node.value:
            if child_key.value == key:
                if not keys:
                    return Range(
                        Position(child_value.start_mark.line, child_value.start_mark.column),
                        Position(child_value.end_mark.line, child_value.end_mark.column),
                    )
                else:
                    return _from_node(child_value).get_value(*keys)

        raise MissingKeyError(key)


class YAMLWhereNull(YAMLWhere):
    "Source map calculator for null nodes."

    def get(self, *keys: str | int) -> Range:
        raise UndefinedAccessError("get() is not defined for null nodes")

    def get_key(self, key: str | int, *keys: str | int) -> Range:
        raise UndefinedAccessError("get_key() is not defined for null nodes")

    def get_value(self, key: str | int, *keys: str | int) -> Range:
        raise UndefinedAccessError("get_value() is not defined for null nodes")


@singledispatch
def _from_node(node: Node) -> YAMLWhere:
    "Construct a YAMLWhere based on the type of the node."
    raise UnsupportedNodeTypeError(
        f"Unsupported node type {type(node).__name__}"
    )  # pragma: no cover


@_from_node.register(ScalarNode)
def _(node):
    return YAMLWhereScalar(node)


@_from_node.register(MappingNode)
def _(node):
    return YAMLWhereMapping(node)


@_from_node.register(SequenceNode)
def _(node):
    return YAMLWhereSequence(node)


@_from_node.register(type(None))
def _(node):
    return YAMLWhereNull(node)




import pytest
from yaml_where.yaml_where import YAMLWhereMapping
from helpers import clean_yaml, rng
from yaml_where import YAMLWhere
from yaml_where.exceptions import MissingKeyError, UndefinedAccessError


def test_get_top_level():
    source_map = YAMLWhere.from_string("a: 1\nb: 42")
    assert source_map.get("a") == rng(0, 0, 0, 4)
    assert source_map.get("b") == rng(1, 0, 1, 5)


def test_get_nested():
    yaml = """
    a:
        b: 42
        c:
            d: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get("a", "b") == rng(1, 4, 1, 9)
    # assert source_map.get("a", "c") == rng(2, 4, 3, 15)
    # assert source_map.get("a", "c", "d") == rng(3, 8, 3, 15)


def test_get_key_top_level():
    source_map = YAMLWhere.from_string("a: 1\nbb: 42")
    assert source_map.get_key("a") == rng(0, 0, 0, 1)
    assert source_map.get_key("bb") == rng(1, 0, 1, 2)


def test_get_key_nested():
    yaml = """
    a:
        b: 42
        c:
            doo: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get_key("a", "b") == rng(1, 4, 1, 5)
    # assert source_map.get_key("a", "c") == rng(2, 4, 2, 5)
    # assert source_map.get_key("a", "c", "doo") == rng(3, 8, 3, 11)


def test_get_value_top_level():
    source_map = YAMLWhere.from_string("a: 1\nbb: 42")
    assert source_map.get_value("a") == rng(0, 3, 0, 4)
    assert source_map.get_value("bb") == rng(1, 4, 1, 6)


def test_get_value_nested():
    yaml = """
    a:
        b: 42
        c:
            doo: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    assert source_map.get_value("a", "b") == rng(1, 7, 1, 9)
    assert source_map.get_value("a", "c") == rng(3, 8, 3, 17)
    assert source_map.get_value("a", "c", "doo") == rng(3, 13, 3, 17)


def test_get_top_level_missing_key():
    source_map = YAMLWhere.from_string("a: 1\nb: 42")
    with pytest.raises(MissingKeyError):
        source_map.get("c")


def test_get_nested_missing_key():
    yaml = """
    a:
        b: 42
        c:
            d: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(MissingKeyError):
        source_map.get("a", "e")


def test_get_key_top_level_missing_key():
    source_map = YAMLWhere.from_string("a: 1\nbb: 42")
    with pytest.raises(MissingKeyError):
        source_map.get_key("c")


def test_get_key_nested_missing_key():
    yaml = """
    a:
        b: 42
        c:
            doo: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(MissingKeyError):
        source_map.get_key("a", "c", "llama")


def test_get_value_top_level_missing_key():
    source_map = YAMLWhere.from_string("a: 1\nbb: 42")
    with pytest.raises(MissingKeyError):
        source_map.get_value("c")


def test_get_value_nested_missing_key():
    yaml = """
    a:
        b: 42
        c:
            doo: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(UndefinedAccessError):
        source_map.get_value("a", "b", "q")


def test_constructor_checks_node_type():
    with pytest.raises(ValueError):
        YAMLWhereMapping(None)


def test_get_with_no_keys():
    yaml = """
    a:
        b: 42
        c:
            doo: hola
    """
    source_map = YAMLWhere.from_string(clean_yaml(yaml))
    with pytest.raises(UndefinedAccessError):
        source_map.get()


import pytest
from yaml_where.exceptions import MissingKeyError, UndefinedAccessError
from helpers import rng
from yaml_where.yaml_where import YAMLWhere, YAMLWhereScalar


def test_get_top_level_scalar():
    yaml = "hello"
    source_map = YAMLWhere.from_string(yaml)
    assert source_map.get() == rng(0, 0, 0, 5)


def test_no_argument_non_scalar():
    yaml = """
    - a: 1
    - b: 2
    """
    source_map = YAMLWhere.from_string(yaml)
    with pytest.raises(UndefinedAccessError):
        source_map.get()


def test_get_with_key():
    yaml = "hello"
    with pytest.raises(UndefinedAccessError):
        YAMLWhere.from_string(yaml).get("a")


def test_get_key():
    yaml = "hello"
    with pytest.raises(UndefinedAccessError):
        YAMLWhere.from_string(yaml).get_key("a")


def test_constructor_checks_node_type():
    with pytest.raises(ValueError):
        YAMLWhereScalar(None)


"""Basic types for representing positions and ranges in a document."""

from dataclasses import dataclass


@dataclass
class Position:
    "Zero-based indexes describing a position in a document."
    line: int
    column: int


@dataclass
class Range:
    "Half-open range describing a span inside a document."
    start: Position
    end: Position



"""Exceptions for yaml-where.
"""


class YAMLWhereException(Exception):
    "Base for all yaml-where exceptions."


class MissingKeyError(YAMLWhereException, KeyError):
    """A key is missing from a mapping or sequence."""


class UndefinedAccessError(YAMLWhereException):
    """An access is undefined for a given node and key."""


class UnsupportedNodeTypeError(YAMLWhereException):
    """A YAML node type is not supported."""



import pytest
from yaml_where.exceptions import UndefinedAccessError
from yaml_where.yaml_where import YAMLWhere


def test_get():
    yaml = ""
    source_map = YAMLWhere.from_string(yaml)
    with pytest.raises(UndefinedAccessError):
        source_map.get()


def test_get_key():
    yaml = ""
    source_map = YAMLWhere.from_string(yaml)
    with pytest.raises(UndefinedAccessError):
        source_map.get_key(0)


def test_get_value():
    yaml = ""
    source_map = YAMLWhere.from_string(yaml)
    with pytest.raises(UndefinedAccessError):
        source_map.get_value(0)


from helpers import clean_yaml, extent
from yaml_where import YAMLWhere


def test_get_seq_in_map():
    yaml = """
    a: [1, 2]
    b:
        c:
            - 4
            - 5
    """
    yw = YAMLWhere.from_string(clean_yaml(yaml))
    assert yw.get("a", 0) == extent(0, 4, 1)
    assert yw.get("a", 1) == extent(0, 7, 1)
    assert yw.get("b", "c", 0) == extent(3, 10, 1)
    assert yw.get("b", "c", 1) == extent(4, 10, 1)


def test_get_map_in_seq():
    yaml = """
    - a: 1
      b: 2
    - [{c: 3}]
    """
    yw = YAMLWhere.from_string(clean_yaml(yaml))
    assert yw.get(0, "a") == extent(0, 2, 4)
    assert yw.get(0, "b") == extent(1, 2, 4)
    assert yw.get(1, 0, "c") == extent(2, 4, 4)


def test_get_key_map_in_seq():
    yaml = """
    - a: 1
      b: 2
    - [{c: 3}]
    """
    yw = YAMLWhere.from_string(clean_yaml(yaml))
    assert yw.get_key(0, "a") == extent(0, 2, 1)
    assert yw.get_key(0, "b") == extent(1, 2, 1)
    assert yw.get_key(1, 0, "c") == extent(2, 4, 1)


def test_get_value_seq_in_map():
    yaml = """
    a: [1, 2]
    b:
        c:
            - 4
            - 5
    """
    yw = YAMLWhere.from_string(clean_yaml(yaml))
    assert yw.get_value("a", 0) == extent(0, 4, 1)
    assert yw.get_value("a", 1) == extent(0, 7, 1)
    assert yw.get_value("b", "c", 0) == extent(3, 10, 1)
    assert yw.get_value("b", "c", 1) == extent(4, 10, 1)


def test_get_value_map_in_seq():
    yaml = """
    - a: 1
      b: 2
    - [{c: 3}]
    """
    yw = YAMLWhere.from_string(clean_yaml(yaml))
    assert yw.get_value(0, "a") == extent(0, 5, 1)
    assert yw.get_value(0, "b") == extent(1, 5, 1)
    assert yw.get_value(1, 0, "c") == extent(2, 7, 1)
